/*******************************************************************************

        @file FileProxy.d
        
        Copyright (C) 2004 Kris Bell
        
        This software is provided 'as-is', without any express or implied
        warranty. In no event will the authors be held liable for damages
        of any kind arising from the use of this software.
        
        Permission is hereby granted to anyone to use this software for any 
        purpose, including commercial applications, and to alter it and/or 
        redistribute it freely, subject to the following restrictions:
        
        1. The origin of this software must not be misrepresented; you must 
           not claim that you wrote the original software. If you use this 
           software in a product, an acknowledgment within documentation of 
           said product would be appreciated but is not required.

        2. Altered source versions must be plainly marked as such, and must 
           not be misrepresented as being the original software.

        3. This notice may not be removed or altered from any distribution
           of the source.


                        ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


        @version        Initial version, March 2004      
        @author         Kris, Brad Anderson, teqdruid


*******************************************************************************/

module mango.io.FileProxy;

private import  mango.base.System;

private import  mango.io.Utf8,
                mango.io.FilePath,
                mango.io.FileStyle,
                mango.io.Exception;

version (linux)
         private extern (C) int strlen (char *s);

version (Win32)
         private extern (C) int wcslen (wchar *s);

/*******************************************************************************

        Models a generic file. Use this to manipulate files and directories
        in conjunction with FilePath, FileSystem and FileConduit. Doxygen
        has a hard time with D version() statements, so part of this class
        is documented in FileProxy::VersionWin32 instead.
        
*******************************************************************************/

class FileProxy
{
        private FilePath path;

        /***********************************************************************
        
                Construct a FileProxy from the provided FilePath

        ***********************************************************************/
                                  
        this (FilePath path)
        {
                this.path = path;
        }

        /***********************************************************************
        
                Construct a FileProxy from a text string

        ***********************************************************************/

        this (char[] path)
        {
                this (new FilePath (path));
        }

        /***********************************************************************
        
                Return the FilePath associated with this FileProxy

        ***********************************************************************/

        FilePath getPath ()
        {
                return path;
        }               

        /***********************************************************************
        
                Return the name of the associated path

        ***********************************************************************/

        char[] toString ()
        {
                return path.toString();
        }               

        /***********************************************************************
        
                Does this path currently exist?

        ***********************************************************************/

        bool isExisting ()               
        {
                try {
                    getSize();
                    return true;
                    } catch (IOException){}
                return false;
        }            

        /***********************************************************************
        
		List the files contained within the associated path:

		@code
		FileProxy proxy = new FileProxy (".");

		foreach (FilePath path; proxy.toList())
			 Stdout.put(path).cr();
		@endcode

        ***********************************************************************/

        FilePath[] toList ()
        {
                bool accept (FilePath fp) {return true;}

                return toList (&accept);
        }              

        /***********************************************************************

        ***********************************************************************/

        version (Win32)
        {
                private import std.c.windows.windows;

                /***************************************************************

                        Throw an exception using the last known error

                ***************************************************************/

                private void exception ()
                {
                        throw new IOException (path.toString ~ ": " ~ System.error);
                }

                /***************************************************************

                        Get info about this path

                ***************************************************************/

                private uint getInfo (void delegate (WIN32_FIND_DATAW info) dg)
                {
                        WIN32_FIND_DATAW info;

                        HANDLE h = FindFirstFileW (path.toUtf16, &info);
                        if (h == INVALID_HANDLE_VALUE)
                            exception ();

                        if (dg)
                            dg (info);
                        FindClose (h);

                        return info.dwFileAttributes;
                }

                /***************************************************************
                        
                        Return the file length (in bytes)

                ***************************************************************/

                long getSize ()
                {
                        long _size;

                        void size (WIN32_FIND_DATAW info)
                        {
                                _size = (cast(ulong) info.nFileSizeHigh << 32) + 
                                                     info.nFileSizeLow;
                        }

                        getInfo (&size);
                        return _size;
                }

                /***************************************************************

                        Is this file writable?

                ***************************************************************/

                bool isWritable ()               
                {
                        return (getInfo(null) & FILE_ATTRIBUTE_READONLY) == 0;
                }            

                /***************************************************************

                        Is this file really a directory?

                ***************************************************************/

                bool isDirectory ()               
                {
                        return (getInfo(null) & FILE_ATTRIBUTE_DIRECTORY) != 0;
                }            

                /***************************************************************

                        Return the time when the file was last modified

                ***************************************************************/

                long getModifiedTime ()
                {
                        long _time;

                        void time (WIN32_FIND_DATAW info)
                        {
                                _time = (cast(ulong) info.ftLastWriteTime.dwHighDateTime << 32) + 
                                                     info.ftLastWriteTime.dwLowDateTime;
                        }

                        getInfo (&time);
                        return _time;
                }               

                /***************************************************************

                        Remove the file/directory from the file-system

                ***************************************************************/

                void remove ()
                {
                        if (isDirectory ())
                           {
                           if (! RemoveDirectoryW (path.toUtf16))
                                 exception();
                           }
                        else
                           {
                           if (! DeleteFileW (path.toUtf16))
                                 exception();
                           }
                }           

                /***************************************************************

                        Create a new file 

                ***************************************************************/

                FileProxy createFile ()
                {
                        HANDLE h;

                        h = CreateFileW (path.toUtf16, GENERIC_WRITE, 0, null, 
                                         CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL, null);

                        if (h == INVALID_HANDLE_VALUE)
                            exception();

                        if (! CloseHandle (h))
                              exception();

                        return this;
                }

                /***************************************************************

                        Create a new directory

                ***************************************************************/

                FileProxy createDirectory ()
                {
                        if (! CreateDirectoryW (path.toUtf16, null))
                              exception();

                        return this;
                }

                /***************************************************************
                        
                        List the set of children within this directory. See
                        toList() above.

                ***************************************************************/

                FilePath[] toList (bool delegate(FilePath fp) filter)
                {
                        int                     i;
                        wchar[]                 c;
                        HANDLE                  h;
                        FilePath[]              list;
                        WIN32_FIND_DATAW        fileinfo;

                        list = new FilePath[50];

                        c = path.toUtf16 ~ cast(wchar[]) "\\*\0";
                        h = FindFirstFileW (c, &fileinfo);

                        if (h != INVALID_HANDLE_VALUE)
                            try {
                                do {
                                   int len = wcslen (fileinfo.cFileName);

                                   // make a copy of the file name for listing
                                   FilePath fp = new FilePath (Utf8.encode(fileinfo.cFileName [0 .. len]));

                                   if (i >= list.length)
                                      list.length = list.length * 2;

                                   if (filter (fp))
                                      {
                                      list[i] = fp;
                                      ++i;
                                      }
                                   } while (FindNextFileW (h, &fileinfo));
                                } finally {
                                          FindClose (h);
                                          }
                        list.length = i;
                        return list;
                }
        }

        /***********************************************************************

        ***********************************************************************/

        version (linux)
        {
                private import std.c.linux.linux;

                /***************************************************************

                        Throw an exception using the last known error

                ***************************************************************/

                private void exception ()
                {
                        throw new IOException (path.toString ~ ": " ~ System.error);
                }

                /***************************************************************

                        Get info about this path

                ***************************************************************/

                private uint getInfo (void delegate (struct_stat info) dg)
                {
                        struct_stat stats;

                        if (std.c.linux.linux.stat (path.toUtf8, &stats))
                            exception();

                        if (dg)
                            dg (stats);

                        return stats.st_mode;
                }               

                /***************************************************************

                        Return the file length (in bytes)

                ***************************************************************/

                long getSize ()
                {
                        long _size;

                        void size (struct_stat info)
                        {
                                _size = cast(ulong) info.st_size;    // 32 bits only
                        }

                        getInfo (&size);
                        return _size;
                }

                /***************************************************************

                        Is this file writable?

                ***************************************************************/

                bool isWritable ()               
                {
                        return (getInfo(null) & O_RDONLY) == 0;
                }            

                /***************************************************************

                        Is this file really a directory?

                ***************************************************************/

                bool isDirectory ()               
                {
                        return (getInfo(null) & S_IFDIR) != 0;
                }            

                /***************************************************************

                        Return the time when the file was last modified

                ***************************************************************/

                long getModifiedTime ()
                {
                        long _time;

                        void time (struct_stat info)
                        {
                               _time = cast(ulong) info.st_mtime; // seconds since 1/1/1970
                        }

                        getInfo (&time);
                        return _time;
                }               

                /***************************************************************

                        Remove the file/directory from the file-system

                ***************************************************************/

                void remove ()
                {
                        if (isDirectory())
                           {
                           if (std.c.linux.linux.rmdir (path.toUtf8))
                               exception ();
                           }
                        else           
                           if (std.c.stdio.remove (path.toUtf8) == -1)
                               exception ();
                }              

                /***************************************************************

                        Create a new file 

                ***************************************************************/

                FileProxy createFile ()
                {
                        int fd;

                        fd = std.c.linux.linux.open (path.toUtf8, O_CREAT | O_WRONLY | O_TRUNC, 0660);
                        if (fd == -1)
                            exception();

                        if (std.c.linux.linux.close(fd) == -1)
                            exception();

                        return this;
                }              

                /***************************************************************

                        Create a new directory

                ***************************************************************/

                FileProxy createDirectory ()
                {
                        if (std.c.linux.linux.mkdir (path.toUtf8, 0777))
                            exception();

                        return this;
                }

                /***************************************************************
                
                        List the set of children within this directory. See
                        toList() above.

                ***************************************************************/

                FilePath[] toList (bool delegate(FilePath fp) filter)
                {
                        int             i;
                        DIR*            dir;
                        dirent*         entry;
                        FilePath[]      list;

                        dir = opendir (path.toUtf8);
                        if (! dir) 
                              exception();

                        list = new FilePath [50];
                        while ((entry = readdir(dir)) != null)
                              {
                              int len = strlen (entry.d_name);

                              // make a copy of the file name for listing
                              FilePath fp = new FilePath (entry.d_name[0 ..len].dup);

                              if (i >= list.length)
                                  list.length = list.length * 2;

                              if (filter (fp))
                                 {
                                 list[i] = fp;
                                 ++i;
                                 }
                              }  

                        list.length = i;
                        closedir (dir);
                        return list;
                }

        }
}
