/*******************************************************************************

        @file Stdio.d
        
        Copyright (C) 2004 Kris Bell
        
        This software is provided 'as-is', without any express or implied
        warranty. In no event will the authors be held liable for damages
        of any kind arising from the use of this software.
        
        Permission is hereby granted to anyone to use this software for any 
        purpose, including commercial applications, and to alter it and/or 
        redistribute it freely, subject to the following restrictions:
        
        1. The origin of this software must not be misrepresented; you must 
           not claim that you wrote the original software. If you use this 
           software in a product, an acknowledgment within documentation of 
           said product would be appreciated but is not required.

        2. Altered source versions must be plainly marked as such, and must 
           not be misrepresented as being the original software.

        3. This notice may not be removed or altered from any distribution
           of the source.


                        ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

       
        @version        Initial version, March 2004      
        @author         Kris


*******************************************************************************/

module mango.io.Stdio;

private import  mango.io.Buffer,
                mango.io.Reader,
                mango.io.Writer,
                mango.io.Tokenizer,
                mango.io.TextReader,
                mango.io.FileConduit,
                mango.io.ConduitStyle,
                mango.io.DisplayWriter;


/*******************************************************************************

        Standard IO readers and writers. These are at module scope so
        there's less typing involved to access them.

*******************************************************************************/

static IReader   Stdin;
static IWriter   Stdout,
                 Stderr;

/*******************************************************************************

        The ubiquitous console IO support. These are standard FileConduit 
        instances, with Reader/Writer wrappers applied appropriately. Note 
        that the outputs use FlushBuffer to automatically flush data as it 
        is added to the buffer. The basic usage of this module is illustrated 
        below:

	@code
	char[] msg = "on the console";

	Stdout.put ("print ")
	      .put (1)
	      .put (' ')
	      .put ("message ")
	      .put (msg)
	      .cr  ();
	@endcode

	An alternative is to use the C++ stream operator like so:

	@code
	char[] msg = "on the console";

	Stdout << "print "
	       << 1
	       << ' '
	       << "message "
	       << msg
	       << Stdout.newline;
	@endcode

        Since console idioms are based upon FileConduit, you can use them 
	as direct targets for Conduit operations. For example, the code:

        @code
          FileConduit fc = new FileConduit ("myfile.txt", FileStyle.Read);
          Stdio.stdout.copy (fc);
        @endcode

        copies a text file directly to the console. Likewise, you can 
        copy console input directly to a FileConduit or a SocketConduit.
        Input via Stdin is similar in nature, but uses the Token classes
        to isolate and parse each token on an input line:

        @code
        int x;
        Stdout.put ("please input a number: ");
        Stdin.get(x);           // or Stdin >> x;
        @endcode

        Note that Stdin awaits a carriage-return before parsing the input
        into the targets. Note also that the Stdout and Stderr are not written 
        to be thread-safe. As such you may find that output from two threads 
        intersect across each other. If this is a problem you should wrap a 
        synchronized block around the offending entity, like so:

        @code
        synchronized (Stdout)
                      Stdout << "this is " << "'atomic' " << " output".cr();
        @endcode
        
        Alternatively, please consider using the mango.log (Logger) package 
        to provide detailed runtime diagnostics from your application. The 
        functionality exposed there is likely sufficient for most application 
        needs.

        Redirecting the standard IO handles (via a shell) operates as one 
	would expect.

*******************************************************************************/

class Stdio
{
        /***********************************************************************
        
                Standard IO conduits. These are inside the Stdio namespace 
                since I prefer their usage to be explicit.

        ***********************************************************************/

        static FileConduit stdin,
                           stdout,
                           stderr;

        /***********************************************************************
        
                Create a set of Stdio conduits and readers/writers

        ***********************************************************************/

        static this()
        {
                stdin  = new FileConduit (new FileDevice(0, ConduitStyle.Access.Read));
                stdout = new FileConduit (new FileDevice(1, ConduitStyle.Access.Write));
                stderr = new FileConduit (new FileDevice(2, ConduitStyle.Access.Write));

                .Stdin  = new TextReader    (stdin.createBuffer(), new LineTokenizer());
                .Stdout = new DisplayWriter (new FlushBuffer (1024 * 16, stdout));
                .Stderr = new DisplayWriter (new FlushBuffer (1024 * 16, stderr));
        }

        /***********************************************************************
        
                Prohibit instantiation of this class 

        ***********************************************************************/

        private this()
        {
        }
}
